import { createContext, useRef, useState, useEffect, createElement, useContext } from 'react';

function deriveInitial(value, defaultValue) {
  const result = {};

  for (const key in value) {
    if (Array.isArray(value[key])) {
      result[key] = value[key].map(val => val && typeof val === 'object' ? deriveInitial(val, defaultValue) : defaultValue);
    } else if (value[key] && typeof value[key] === 'object') {
      result[key] = deriveInitial(value[key], defaultValue);
    } else {
      result[key] = defaultValue;
    }
  }

  return result;
}

function get(source, key) {
  let path = key.replace(/\[("|')?([^\[\]]+)\1\]/g, '.$2').split('.'),
      index = 0;

  while (source && path.length > index) {
    source = source[path[index++]];
  }

  return source;
}
function set(source, key, value) {
  return setHelper(source, value, key.replace(/\[("|')?([^\[\]]+)\1\]/g, '.$2').split('.'), 0);
}

function setHelper(source, value, pathArray, currentIndex) {
  if (currentIndex >= pathArray.length) return value;
  const continuedPath = setHelper(source && source[pathArray[currentIndex]], value, pathArray, currentIndex + 1);

  if (!source) {
    if (isNaN(pathArray[currentIndex])) {
      return {
        [pathArray[currentIndex]]: continuedPath
      };
    }

    const result = [];
    result[pathArray[currentIndex]] = continuedPath;
    return result;
  }

  if (Array.isArray(source)) {
    source[pathArray[currentIndex]] = continuedPath;
    return source;
  }

  return Object.assign({}, source, {
    [pathArray[currentIndex]]: continuedPath
  });
}

const EMPTY_OBJ = {};
const formContext = createContext(EMPTY_OBJ);

function Form({
  children,
  enableReinitialize,
  initialErrors,
  initialValues,
  onSubmit,
  noForm,
  validate,
  onError,
  onSuccess,
  shouldSubmitWhenInvalid,
  validateOnBlur,
  validateOnChange,
  ...formProps
}) {
  const fieldValidators = useRef([]);
  const isDirty = useRef(false);
  const {
    0: values,
    1: setValues
  } = useState(initialValues || EMPTY_OBJ);
  const {
    0: touched,
    1: setTouched
  } = useState(initialErrors && (() => deriveInitial(initialErrors, true)) || EMPTY_OBJ);
  const {
    0: errors,
    1: setErrors
  } = useState(initialErrors || EMPTY_OBJ);
  const submittingState = useState(false);
  const formErrorState = useState();

  const validateForm = () => {
    let validationErrors = validate && validate(values) || EMPTY_OBJ;
    fieldValidators.current.some(tuple => {
      const error = tuple[1](get(values, tuple[0]));

      if (error) {
        validationErrors = set(validationErrors, tuple[0], error);
      }
    });
    if (fieldValidators.current.length) validationErrors = JSON.parse(JSON.stringify(validationErrors));

    if (validationErrors !== errors && JSON.stringify(errors) !== JSON.stringify(validationErrors)) {
      setErrors(validationErrors);
    }

    return validationErrors;
  };

  const resetForm = () => {
    isDirty.current = false;
    setValues(initialValues || EMPTY_OBJ);

    if (initialErrors) {
      setTouched(deriveInitial(initialErrors, true));
      setErrors(initialErrors);
    }
  };

  const handleSubmit = event => {
    if (event && event.preventDefault) event.preventDefault();
    submittingState[1](true);
    const fieldErrors = validateForm();
    setTouched(deriveInitial(fieldErrors, true));

    if (!shouldSubmitWhenInvalid && Object.keys(fieldErrors).length > 0) {
      return submittingState[1](false);
    }

    return new Promise(resolve => resolve(onSubmit(values, {
      setErrors: submitErrors => {
        setErrors(submitErrors);
      },
      setFormError: err => {
        formErrorState[1](err);
      }
    }))).then(result => {
      submittingState[1](false);
      if (onSuccess) onSuccess(result, {
        resetForm
      });
    }, err => {
      submittingState[1](false);
      if (onError) onError(err, {
        setErrors: submitErrors => {
          setErrors(submitErrors);
        },
        setFormError: err => {
          formErrorState[1](err);
        }
      });
    });
  };

  useEffect(() => {
    if (enableReinitialize) resetForm();
  }, [initialValues]);
  useEffect(() => {
    if ((validateOnBlur === undefined || validateOnChange || validateOnBlur) && isDirty.current) {
      validateForm();
    }
  }, [validateOnBlur === undefined ? touched : validateOnBlur && touched, validateOnChange && values, isDirty.current]);

  const change = (fieldId, value) => {
    isDirty.current = true;
    setValues(state => set(state, fieldId, value));
  };

  const toRender = typeof children === 'function' ? children({
    change,
    formError: formErrorState[0],
    isDirty: isDirty.current,
    isSubmitting: submittingState[0],
    handleSubmit,
    resetForm
  }) : children;
  return createElement(formContext.Provider, {
    value: {
      errors: errors,
      formError: formErrorState[0],
      isDirty: isDirty.current,
      isSubmitting: submittingState[0],
      resetForm,
      setFieldError: (fieldId, error) => {
        setErrors(state => set(state, fieldId, error));
      },
      setFieldTouched: (fieldId, value) => {
        setTouched(state => set(state, fieldId, value));
      },
      setFieldValue: change,
      submit: handleSubmit,
      touched: touched,
      validate: validateForm,
      values,
      _fieldValidators: fieldValidators.current
    }
  }, noForm ? toRender : createElement("form", Object.assign({
    onSubmit: handleSubmit
  }, formProps), toRender));
}

function useField(fieldId, validate) {
  if (process.env.NODE_ENV !== 'production' && (!fieldId || typeof fieldId !== 'string')) {
    throw new Error('The Field needs a valid "fieldId" property to function correctly.');
  }

  const ctx = useContext(formContext);
  const isMounted = useRef(false);
  useEffect(() => {
    const tuple = [fieldId, validate];

    if (validate) {
      ctx._fieldValidators.push(tuple);

      if (isMounted.current) {
        const newError = validate(get(ctx.values, fieldId));

        if (newError !== get(ctx.errors, fieldId)) {
          ctx.setFieldError(fieldId, newError);
        }
      }
    } else if (get(ctx.errors, fieldId)) {
      ctx.setFieldError(fieldId, undefined);
    }

    isMounted.current = true;
    return () => {
      if (validate) {
        ctx._fieldValidators.splice(ctx._fieldValidators.indexOf(tuple), 1);
      }
    };
  }, [fieldId, validate]);
  return [{
    onBlur: () => {
      ctx.setFieldTouched(fieldId, true);
    },
    onChange: value => {
      ctx.setFieldValue(fieldId, value);
    },
    onFocus: () => {
      ctx.setFieldTouched(fieldId, false);
    }
  }, {
    error: get(ctx.errors, fieldId),
    touched: get(ctx.touched, fieldId),
    value: get(ctx.values, fieldId) || ''
  }];
}

function useFieldArray(fieldId) {
  if (process.env.NODE_ENV !== 'production' && (!fieldId || typeof fieldId !== 'string')) {
    throw new Error('The FieldArray needs a valid "fieldId" property to function correctly.');
  }

  const ctx = useContext(formContext);
  const value = get(ctx.values, fieldId) || [];
  return [{
    add: element => {
      ctx.setFieldValue(fieldId, [...value, element]);
    },
    insert: (at, element) => {
      const touched = get(ctx.touched, fieldId) || [];
      const errors = get(ctx.errors, fieldId);
      value.splice(at, 0, element);
      touched.splice(at, 0, false);
      ctx.setFieldValue(fieldId, value);
      ctx.setFieldTouched(fieldId, touched);

      if (errors) {
        errors.splice(at, 0, undefined);
        ctx.setFieldError(fieldId, errors);
      }
    },
    remove: index => {
      const touched = get(ctx.touched, fieldId) || [];
      const errors = get(ctx.errors, fieldId);
      value.splice(index, 1);
      touched.splice(index, 1);
      ctx.setFieldValue(fieldId, value);
      ctx.setFieldTouched(fieldId, touched);

      if (errors) {
        errors.splice(index, 1);
        ctx.setFieldError(fieldId, errors);
      }
    }
  }, {
    error: get(ctx.errors, fieldId),
    value
  }];
}

function useError(fieldId) {
  if (process.env.NODE_ENV !== 'production' && (!fieldId || typeof fieldId !== 'string')) {
    throw new Error('The Error needs a valid "fieldId" property to function correctly.');
  }

  return get(useContext(formContext).errors, fieldId);
}

var useFormConnect = (() => useContext(formContext));

function OptionsContainer(options) {
  let isMounted = false;
  return function FormOuterWrapper(Component) {
    return function FormWrapper(props) {
      const initialValuesState = useState(() => options.mapPropsToValues && options.mapPropsToValues(props));
      useEffect(() => {
        if (options.enableReinitialize && options.mapPropsToValues && isMounted) initialValuesState[1](options.mapPropsToValues(props));
        isMounted = true;
      }, options.enableReinitialize ? Object.values(props) : []);
      return createElement(Form, Object.assign({}, options, {
        initialValues: initialValuesState[0] || options.initialValues,
        noForm: true,
        validateOnBlur: options.validateOnBlur === undefined ? false : options.validateOnBlur
      }), form => createElement(Component, Object.assign({}, form, props)));
    };
  };
}

function useSpy(fieldId, cb) {
  const ctx = useContext(formContext);
  const isMounted = useRef(false);
  const prev = useRef();
  const value = get(ctx.values, fieldId);
  useEffect(() => {
    if (isMounted.current && cb && prev.current !== value) cb(value, ctx);
    isMounted.current = true;
    prev.current = value;
  }, [value]);
  return {
    error: get(ctx.errors, fieldId),
    touched: get(ctx.touched, fieldId),
    value
  };
}

export { OptionsContainer as Form, Form as HookedForm, useError, useField, useFieldArray, useFormConnect, useSpy };
//# sourceMappingURL=hooked-form.modern.js.map
